"use strict";
/**
Copyright (c) 2014 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.
**/

require("../../base/range.js");
require("../../extras/chrome/chrome_user_friendly_category_driver.js");
require("./chrome_process_helper.js");

'use strict';

global.tr.exportTo('tr.model.helpers', function () {
  function ChromeRendererHelper(modelHelper, process) {
    tr.model.helpers.ChromeProcessHelper.call(this, modelHelper, process);
    this.mainThread_ = process.findAtMostOneThreadNamed('CrRendererMain');
    this.compositorThread_ = process.findAtMostOneThreadNamed('Compositor');
    this.rasterWorkerThreads_ = process.findAllThreadsMatching(function (t) {
      if (t.name === undefined) return false;
      if (t.name.indexOf('CompositorTileWorker') === 0) return true;
      if (t.name.indexOf('CompositorRasterWorker') === 0) return true;
      return false;
    });

    this.isChromeTracingUI_ = process.labels !== undefined && process.labels.length === 1 && process.labels[0] === 'chrome://tracing';
    if (!process.name) process.name = ChromeRendererHelper.PROCESS_NAME;
  }

  ChromeRendererHelper.PROCESS_NAME = 'Renderer';

  // Returns true if there is either a main thread or a compositor thread.
  ChromeRendererHelper.isRenderProcess = function (process) {
    if (process.findAtMostOneThreadNamed('CrRendererMain')) return true;
    if (process.findAtMostOneThreadNamed('Compositor')) return true;
    return false;
  };

  ChromeRendererHelper.prototype = {
    __proto__: tr.model.helpers.ChromeProcessHelper.prototype,

    // May be undefined.
    get mainThread() {
      return this.mainThread_;
    },

    // May be undefined.
    get compositorThread() {
      return this.compositorThread_;
    },

    // May be empty.
    get rasterWorkerThreads() {
      return this.rasterWorkerThreads_;
    },

    get isChromeTracingUI() {
      return this.isChromeTracingUI_;
    },

    /**
    * Generate a breakdown that attributes where time goes between |start| &
    * |end| on renderer thread.
    *
    * @param {number} start
    * @param {number} end
    * @return {Object} A time breakdown object whose every key is a chrome
    * userfriendly title & values are an object that show the total spent
    * between |start| & |end|, and the list of event labels of the group and
    * their total time between |start| & |end|.
    *
    * Example:
    *   {layout: {
    *        total: 100,
    *        events: {'FrameView::performPreLayoutTasks': 20,..}},
    *    v8_runtime: {
    *        total: 500,
    *        events: {'String::NewExternalTwoByte': 0.5,..}},
    *    ...
    *    }
    *
    *
    */
    generateTimeBreakdownTree: function (start, end) {
      if (this.mainThread === null) return;
      var breakdownMap = {};
      var range = tr.b.Range.fromExplicitRange(start, end);
      for (var title of tr.e.chrome.ChromeUserFriendlyCategoryDriver.ALL_TITLES) {
        breakdownMap[title] = { total: 0, events: {} };
      }
      breakdownMap['idle'] = { total: 0, events: {} };
      var totalIdleTime = end - start;
      for (var event of this.mainThread.getDescendantEvents()) {
        if (!range.intersectsExplicitRangeExclusive(event.start, event.end)) continue;
        if (event.selfTime === undefined) continue;
        var title = tr.e.chrome.ChromeUserFriendlyCategoryDriver.fromEvent(event);
        var wallTimeIntersectionRatio = 0;
        if (event.duration > 0) {
          wallTimeIntersectionRatio = range.findExplicitIntersectionDuration(event.start, event.end) / event.duration;
        }
        var v8Runtime = event.args['runtime-call-stat'];
        if (v8Runtime !== undefined) {
          try {
            var v8RuntimeObject = JSON.parse(v8Runtime);
            for (var runtimeCall in v8RuntimeObject) {
              if (v8RuntimeObject[runtimeCall].length == 2) {
                if (breakdownMap['v8_runtime'].events[runtimeCall] === undefined) {
                  breakdownMap['v8_runtime'].events[runtimeCall] = 0;
                }
                // V8 Runtime Call Stats data is in us, while the
                // breakdown tree timing is in ms.
                var runtimeTime = v8RuntimeObject[runtimeCall][1] * wallTimeIntersectionRatio / 1000;
                breakdownMap['v8_runtime'].total += runtimeTime;
                breakdownMap['v8_runtime'].events[runtimeCall] += runtimeTime;
              }
            }
          } catch (e) {
            console.warn(e);
          }
        }
        //        [     Slice 1       ]   [      Slice  2   ]   [    Slice 3   ]
        //            [  Slice 4    ]                             [ Slice 5 ]
        //              [ Slice 6 ]                                  |
        //                 |                                         |
        //                 |                                         |
        //                 v                                         v
        //                start                                     end
        //
        // For the case where the |start| or |end| overlapped with some existing
        // slice (see above diagram), we approximate the overlapped self-time
        // by multiplying the ratio of overlapped wall time to the self-time.
        // There should be way to compute the exact number, but in practice,
        // this should rarely happen, and when it does, the overlapped range
        // is relative small so that using approximation here should be good
        // enough.
        var approximatedSelfTimeContribution = event.selfTime * wallTimeIntersectionRatio;
        breakdownMap[title].total += approximatedSelfTimeContribution;
        if (breakdownMap[title].events[event.title] === undefined) breakdownMap[title].events[event.title] = 0;
        breakdownMap[title].events[event.title] += approximatedSelfTimeContribution;
        totalIdleTime -= approximatedSelfTimeContribution;
      }
      breakdownMap['idle'].total = totalIdleTime;
      return breakdownMap;
    }

  };

  return {
    ChromeRendererHelper: ChromeRendererHelper
  };
});